#include "tp.c"
#include "vmsdl.c"

typedef struct tp_code_r { unsigned char i,a,b,c; } tp_code_r;
typedef struct tp_code_s { char value[4]; } tp_code_s;
typedef struct tp_code_n { float value; } tp_code_n;
typedef union tp_code {
    unsigned char i;
    tp_code_r regs;
    tp_code_s string;
    tp_code_n number;
} tp_code;

typedef struct tp_frame {
    tp_code *codes;
    int cur;
    int jmp;
    tp_obj *regs;
    tp_obj *ret_dest;
    tp_obj fname;
    tp_obj name;
    tp_obj line;
    tp_obj globals;
    int lineno;
} tp_frame;

#define TP_FRAMES 64
#define TP_REGS_PER_FRAME 256
#define TP_REGS (TP_REGS_PER_FRAME*TP_FRAMES)
typedef struct tp_vm {
    tp_obj builtins;
    tp_obj modules;
    tp_frame frames[TP_FRAMES];
    tp_obj regs[TP_REGS];
    jmp_buf buf;
    tp_obj ex;
    int cur;
} tp_vm;

tp_vm *vm_init(tp_obj builtins) {
    tp_vm * vm = tp_malloc(sizeof(tp_vm));
    vm->builtins = builtins;
    vm->modules = tp_dict(0,0);
    tp_set(builtins,tp_string("MODULES"),vm->modules);
    vm->cur = -1;
    vm->ex = None;
    int i;
    for (i=0; i<TP_REGS; i++) { vm->regs[i] = None; }
    return vm;
}


tp_frame vm_frame(tp_obj globals,tp_code *codes,tp_obj *regs,tp_obj *ret_dest) {
    tp_frame f;
    f.globals = globals;
    f.codes = codes;
    f.cur = 0;
    f.jmp = 0;
    f.regs = regs;
    f.ret_dest = ret_dest;
    f.lineno = 0;
    f.line = tp_string("");
    f.name = tp_string("?");
    f.fname = tp_string("?");
    return f;
}

typedef enum vm_instructions {
    VM_EOF,VM_ADD,VM_SUB,VM_MUL,VM_DIV,VM_POW,VM_AND,VM_OR,VM_CMP,VM_GET,VM_SET,
    VM_NUMBER,VM_STRING,VM_GGET,VM_GSET,VM_MOVE,VM_DEF,VM_PASS,VM_JUMP,VM_CALL,
    VM_RETURN,VM_IF,VM_DEBUG,VM_EQ,VM_LE,VM_LT,VM_DICT,VM_LIST,VM_NONE,VM_LEN,
    VM_LINE,VM_PARAMS,VM_IGET,VM_FILE,VM_NAME,VM_NE,VM_HAS,VM_RAISE,VM_SETJMP,
    VM_MOD,VM_LSH,VM_RSH,VM_ITER,VM_DEL,
    VM_TOTAL
} vm_instructions;

char *vm_strings[VM_TOTAL] = {
    "EOF","ADD","SUB","MUL","DIV","POW","AND","OR","CMP","GET","SET","NUM",
    "STR","GGET","GSET","MOVE","DEF","PASS","JUMP","CALL","RETURN","IF","DEBUG",
    "EQ","LE","LT","DICT","LIST","NONE","LEN","LINE","PARAMS","IGET","FILE",
    "NAME","NE","HAS","RAISE","SETJMP","MOD","LSH","RSH","ITER","DEL",
};

#define VA ((int)e.regs.a)
#define VB ((int)e.regs.b)
#define VC ((int)e.regs.c)
#define RA regs[e.regs.a]
#define RB regs[e.regs.b]
#define RC regs[e.regs.c]
#define R0 regs[0]
#define NA regs[e.regs.a].number.value
#define NB regs[e.regs.b].number.value
#define NC regs[e.regs.c].number.value
#define NUM(n) tp_number(n)
#define UVBC (unsigned short)(((VB<<8)+VC))
#define SVBC (short)(((VB<<8)+VC))

static tp_vm *VM;

void _vm_raise(tp_obj e) {
    VM->ex = e;
    longjmp(VM->buf,1);
}

void vm_print_stack() {
    printf("\n");
    int i;
    for (i=0; i<=VM->cur; i++) {
        printf("File \"%s\", line %d, in %s\n  %s\n",
            STR(VM->frames[i].fname),VM->frames[i].lineno,STR(VM->frames[i].name),STR(VM->frames[i].line));
    }
    printf("\nException:\n");
    printf("%s\n",tp_cstring(tp_str(VM->ex)));
}



void vm_handle() {
    int i;
    for (i=VM->cur; i>=0; i--) {
        if (VM->frames[i].jmp) { break; }
    }
    if (i >= 0) {
        VM->cur = i;
        VM->frames[i].cur = VM->frames[i].jmp;
        VM->frames[i].jmp = 0;
        return;
    }

    vm_print_stack();
    exit(-1);
}

void _vm_call(tp_vm *vm,tp_obj *dest, tp_obj fnc, tp_obj params) {
    if (fnc.type == TP_FNC) {
        *dest = _tcall(fnc,params);
        return;
    }
    if (fnc.type == TP_DEF) {
        tp_frame *f = &vm->frames[vm->cur];
        vm->cur += 1;
        vm->frames[vm->cur] = vm_frame(*(tp_obj*)fnc.def.globals,fnc.def.value,&vm->regs[vm->cur*TP_REGS_PER_FRAME],dest);
        if (fnc.def.self) {
            vm->frames[vm->cur].regs[0] = params;
            _tp_list_insert(params.list.value,0,fnc.def.self);
        } else {
            vm->frames[vm->cur].regs[0] = params;
        }
        return;
    }
    print(tp_params(1,(tp_obj[]){fnc}));
    vm_raise("tp_call: %s is not callable",STR(fnc));
}

void vm_return(tp_vm *vm, tp_obj v) {
    tp_obj *dest = vm->frames[vm->cur].ret_dest;
    if (dest) { *dest = v; }
    vm->cur -= 1;
}


inline int tp_step(tp_vm *vm) {
    tp_frame *f = &vm->frames[vm->cur];
    tp_code e = f->codes[f->cur];
    tp_obj *regs = f->regs;
    switch (e.i) {
        case VM_EOF:
            if (vm->cur) { vm_return(vm,None); return; }
            return -1;
            break;
        case VM_ADD: RA = tp_add(RB,RC); break;
        case VM_SUB: RA = tp_sub(RB,RC); break;
        case VM_MUL: RA = tp_mul(RB,RC); break;
        case VM_DIV: RA = tp_div(RB,RC); break;
        case VM_POW: RA = tp_pow(RB,RC); break;
        case VM_AND: RA = tp_and(RB,RC); break;
        case VM_OR:  RA = tp_or(RB,RC); break;
        case VM_MOD:  RA = tp_mod(RB,RC); break;
        case VM_LSH:  RA = tp_lsh(RB,RC); break;
        case VM_RSH:  RA = tp_rsh(RB,RC); break;
        case VM_CMP: RA = tp_cmp(RB,RC); break;
        case VM_NE: RA = tp_number(tp_cmp(RB,RC).number.value!=0); break;
        case VM_EQ: RA = tp_number(tp_cmp(RB,RC).number.value==0); break;
        case VM_LE: RA = tp_number(tp_cmp(RB,RC).number.value<=0); break;
        case VM_LT: RA = tp_number(tp_cmp(RB,RC).number.value<0); break;
        case VM_PASS: break;
        case VM_IF: if (RA.number.value) { f->cur += 1; } break;
        case VM_GET: RA = tp_get(RB,RC); break;
        case VM_ITER: RA = tp_iter(RB,RC); break;
        case VM_HAS: RA = tp_has(RB,RC); break;
        case VM_IGET: tp_iget(&RA,RB,RC); break;
        case VM_SET: tp_set(RA,RB,RC); break;
        case VM_DEL: tp_del(RA,RB); break;
        case VM_MOVE: RA = RB; break;
        case VM_NUMBER: RA = tp_number(f->codes[++f->cur].number.value); break;
        case VM_STRING:
            RA = tp_string_n(f->codes[f->cur+1].string.value,UVBC);
            f->cur += (UVBC/4)+1;
            break;
        case VM_DICT: RA = tp_dict(VC,&RB); break;
        case VM_LIST: RA = tp_list(VC,&RB); break;
        case VM_PARAMS: RA = tp_params(VC,&RB); break;
        case VM_LEN: RA = tp_len(RB); break;
        case VM_JUMP: f->cur += SVBC; return; break;
        case VM_SETJMP: f->jmp = f->cur+SVBC; break;
        case VM_CALL: _vm_call(vm,&RA,RB,RC); break;
        case VM_GGET:
            if (tp_has(f->globals,RB).number.value) {
                RA = tp_get(f->globals,RB);
            } else { RA = tp_get(vm->builtins,RB); }
            break;
        case VM_GSET: tp_set(f->globals,RA,RB); break;
        case VM_DEF:
            RA = tp_def(f->codes[f->cur+1].string.value,tp_ptr(f->globals));
            f->cur = UVBC;
            return;
            break;
        case VM_RETURN: vm_return(vm,RA); return; break;
        case VM_RAISE: _vm_raise(RA); return; break;
        case VM_DEBUG:
            print(tp_params(2,(tp_obj[]){tp_string("DEBUG:"),RA}));
            break;
        case VM_NONE: RA = None; break;
        case VM_LINE:
            f->line = tp_string(f->codes[f->cur+1].string.value);
            f->cur += VA; f->lineno = UVBC; break;
        case VM_FILE: f->fname = RA; break;
        case VM_NAME: f->name = RA; break;
        default: vm_raise("tp_step: invalid instruction %d",e.i); break;
    }
    f->cur += 1;
    return 0;
}

tp_obj _import(char *s,tp_obj name) {
    if (!file_exists(s)) {
        vm_raise("import(%s) - file does not exist",s);
    }
    FILE *f;
    long l = FileSize(s);
    f = fopen(s,"r");
    tp_code *codes = malloc(l);
    fread(codes,1,l,f);
    fclose(f);

    tp_vm *vm = VM;
    vm->cur += 1;
    tp_obj g = tp_dict(2,(tp_obj[]){tp_string("__name__"),name});
    vm->frames[vm->cur] =   vm_frame(g,codes,&vm->regs[vm->cur*TP_REGS_PER_FRAME],0);
    return g;
}

void vm_run_1() {
    int step = 0;
    int cur = VM->cur;
    if (setjmp(VM->buf)) { vm_handle(); }
    while (VM->cur >= cur && tp_step(VM) != -1);
    VM->cur = cur-1;
}


tp_obj vm_bmod(char *name, void *codes) {
    tp_vm *vm = VM;
    vm->cur += 1;
    tp_obj g = tp_dict(2,(tp_obj[]){tp_string("__name__"),tp_string(name)});
    tp_set(VM->modules,tp_string(name),g);
    vm->frames[vm->cur] =   vm_frame(g,codes,&vm->regs[vm->cur*TP_REGS_PER_FRAME],0);
    vm_run_1();
    return g;
}

tp_obj tp_exec(tp_obj params) {
    P_BEGIN(params);
    tp_obj code = P_OBJ("code");
    tp_obj globals = P_OBJ("globals");
    P_END;
    tp_vm *vm = VM;
    vm->cur += 1;
    vm->frames[vm->cur] = vm_frame(globals,(void*)code.string.buf,&vm->regs[vm->cur*TP_REGS_PER_FRAME],0);
    return None;
}

tp_obj vm_call(tp_obj fnc, tp_obj params) {
    tp_obj r;
    _vm_call(VM,&r,fnc,params);
    vm_run_1();
    return r;
}

tp_obj import(tp_obj params) {
    P_BEGIN(params);
    tp_obj mod = P_OBJ("module")
    P_END;

    if (tp_has(VM->modules,mod).number.value) {
        return tp_get(VM->modules,mod);
    }

    char *s = tp_cstring(tp_add(mod,tp_string(".tpc")));
    tp_obj r = _import(s,mod);
    tp_set(VM->modules,mod,r);
    return r;
}

#define VM_B(a,b) tp_set(ctx,tp_string(a),tp_fnc(b))

tp_obj vm_builtins() {
    tp_obj ctx = tp_dict(0,0);
    VM_B("print",print); VM_B("range",range); VM_B("min",min);
    VM_B("max",max); VM_B("bind",bind); VM_B("copy",tp_copy);
    VM_B("import",import); VM_B("len",len); VM_B("assert",tp_assert);
    VM_B("str",tp_str2); VM_B("float",tp_float); VM_B("system",tp_system);
    VM_B("istype",tp_istype); VM_B("chr",tp_chr); VM_B("save",tp_save);
    VM_B("load",tp_load); VM_B("fpack",tp_fpack); VM_B("abs",tp_abs);
    VM_B("int",tp_int);
    VM_B("exec",tp_exec);
    VM_B("exists",tp_exists); VM_B("mtime",tp_mtime);
    VM_B("number",tp_float);
    // ord, round, hex, unlink
    sdl_init(ctx);
    return ctx;
}

void vm_run() {
    int step = 0;
    if (setjmp(VM->buf)) { vm_handle(); }
    while (tp_step(VM) != -1) {
    }
}

void vm_signal(int v) {
    vm_raise("%s",strsignal(v));
}

void vm_args(tp_obj ctx, int argc, char *argv[]) {
    tp_obj self = tp_list(0,0);
    int n;
    for (n=1; n<argc; n++) {
        _tp_list_append(self.list.value,tp_ptr(tp_string(argv[n])));
    }
    tp_set(ctx,tp_string("ARGV"),self);
}

#ifndef TP_MAIN
int main(int argc, char *argv[]) {
    tp_obj ctx = vm_builtins();
    vm_args(ctx,argc,argv);
    VM = vm_init(ctx);
    if (setjmp(VM->buf)) { vm_handle(); }
//     signal(SIGSEGV, vm_signal);
    _import(argv[1],tp_string("__main__"));
    vm_run();
    return(0);
}
#endif

//
